from io import BytesIO
from itertools import groupby
from math import radians
from typing import Optional

import pytest
from OCP.BRepPrimAPI import BRepPrimAPI_MakeCylinder
from OCP.gp import gp_Ax2, gp_Dir, gp_Pnt
from OCP.TopoDS import TopoDS_Shape

from ocpsvg.hlr import HiddenLineRenderer, Padding, basic_style, write_svg
from ocpsvg.ocp import edge_from_curve, segment_curve


def test_edge_order_ortho():
    S0 = cylinder(1, 3, x=0, y=0)
    S1 = cylinder(1, 3, x=4, y=4)
    S2 = cylinder(1, 3, x=8, y=8)

    render = HiddenLineRenderer.Orthographic(gp_Dir(-1, -1, -1))([S0, S1, S2])
    assert [i for i, _ in groupby(e.shape_index for e in render)] == [0, 1, 2]

    render = HiddenLineRenderer.Orthographic(gp_Dir(+1, +1, -1))([S0, S1, S2])
    assert [i for i, _ in groupby(e.shape_index for e in render)] == [2, 1, 0]


def test_edge_order_perspective():
    S0 = cylinder(1, 3, x=0, y=0)
    S1 = cylinder(1, 3, x=4, y=4)
    S2 = cylinder(1, 3, x=8, y=8)

    render = HiddenLineRenderer.Perspective(gp_Pnt(+1, +1, +1))([S0, S1, S2])
    assert [i for i, _ in groupby(e.shape_index for e in render)] == [0, 1, 2]

    render = HiddenLineRenderer.Perspective(gp_Pnt(-1, -1, +1))([S0, S1, S2])
    assert [i for i, _ in groupby(e.shape_index for e in render)] == [2, 1, 0]


def test_to_svg():
    shapes = [
        cylinder(1, 3, x=0, y=0),
        cylinder(1, 3, x=4, y=4),
        cylinder(1, 3, x=8, y=8),
    ]

    render = HiddenLineRenderer.Orthographic(gp_Dir(-1, -1, -1))(shapes)
    svg = render.to_svg()
    path_elements = svg.findall("*/path")

    assert len(path_elements) == len(list(render))

    for element, edge in zip(path_elements, render):
        assert f"s{edge.shape_index}" in element.attrib["class"]


def test_to_svg_background():
    shapes = [
        cylinder(1, 3, x=0, y=0),
    ]
    render = HiddenLineRenderer.Orthographic(gp_Dir(-1, -1, -1))(shapes)

    svg = render.to_svg(css_style={})
    assert not svg.findall('rect[@id="background"]')

    svg = render.to_svg(background=True)
    assert svg.findall('rect[@id="background"]')

    svg = render.to_svg(css_style={"#background": {"color": "white"}})
    assert svg.findall('rect[@id="background"]')


def test_basic_style_color():
    css = basic_style(color=((1, 0, 0), (0, 1, 0)))
    assert css["path"]["stroke"] == "#ff0000"
    assert css[".hidden"]["stroke"] == "#00ff00"

    css = basic_style(color=(0, 0, 0))
    assert css["path"]["stroke"] == "#000000"
    assert css[".hidden"]["stroke"] == "#000000"


def test_basic_style_shape_color():
    css = basic_style(
        shape_colors={
            0: ((1, 0, 0), (0, 1, 0)),
            2: ((1, 0, 0), (0, 0, 1)),
        }
    )
    assert css[".s0"]["stroke"] == "#ff0000"
    assert css[".s0.hidden"]["stroke"] == "#00ff00"
    assert css[".s2"]["stroke"] == "#ff0000"
    assert css[".s2.hidden"]["stroke"] == "#0000ff"


def test_basic_style_background_color():
    css = basic_style(background_color=None)
    assert "#background" not in css

    css = basic_style(background_color=(1, 1, 1))
    assert css["#background"]["fill"] == "#ffffff"


# def test_viewport():
#     shapes = [edge_from_curve(segment_curve(gp_Pnt(1, 2, 0), gp_Pnt(11, 22, 3)))]
#     render = HiddenLineRenderer.Orthographic(gp_Dir(-1, 0, 0))(shapes)
#     svg = render.to_svg(width=10, padding=2, css_style={})


def test_write_svg():
    shapes = [edge_from_curve(segment_curve(gp_Pnt(0, 0, 0), gp_Pnt(1, 2, 3)))]
    render = HiddenLineRenderer.Orthographic(gp_Dir(-1, 0, 0))(shapes)

    svg = render.to_svg(width=10, padding=2, css_style={})
    buf = BytesIO()
    write_svg(svg, buf)
    assert buf.getvalue() == (
        b"<?xml version='1.0' encoding='utf-8'?>\n<svg xmlns=\"http://www.w3.org"
        b'/2000/svg" viewBox="-2 -2 10 13" width="10">\n  <g transform="scale(3 -3)'
        b' translate(0 -3)">\n    <path id="e0" d="M0,0 L2,3" class="s0 sharp" />\n '
        b" </g>\n</svg>"
    )


@pytest.mark.parametrize(
    "params, expected",
    [
        ((None, None, 0), ("2 3 10 10", None, None)),
        ((200, None, 0), ("40 60 200 200", "200", None)),
        ((None, 200, 0), ("40 60 200 200", None, "200")),
        ((400, 200, 0), ("40 60 200 200", "400", "200")),
        ((200, 400, 0), ("40 60 200 200", "200", "400")),
        ((200, None, 10), ("26 44 200 200", "200", None)),
        ((200, None, (20,)), ("12 28 200 200", "200", None)),
        ((200, None, (10, 20)), ("12 38 200 180", "200", None)),
        ((200, None, (10, 20, 30)), ("12 18 200 200", "200", None)),
        ((200, None, (10, 20, 30, 40)), ("-12 12 200 180", "200", None)),
    ],
)
def test_size_and_padding(
    params: tuple[Optional[float], Optional[float], Padding],
    expected: tuple[str, Optional[str], Optional[str]],
):
    shapes = [edge_from_curve(segment_curve(gp_Pnt(1, 2, 3), gp_Pnt(11, 12, 13)))]
    render = HiddenLineRenderer.Orthographic(gp_Dir(-1, 0, 0))(shapes)
    width, height, padding = params
    svg = render.to_svg(width=width, height=height, padding=padding)
    viewbox = svg.getroot().attrib["viewBox"]
    width = svg.getroot().attrib.get("width")
    height = svg.getroot().attrib.get("height")
    assert (viewbox, width, height) == expected


def cylinder(
    radius: float, height: float, x: float = 0, y: float = 0, z: float = 0
) -> TopoDS_Shape:
    ax = gp_Ax2(gp_Pnt(x, y, z), gp_Dir(0, 0, 1))
    return BRepPrimAPI_MakeCylinder(ax, radius, height, radians(360)).Solid()



import logging
import pathlib
from itertools import chain
from math import degrees, pi
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    Literal,
    NamedTuple,
    Optional,
    Sequence,
    TextIO,
    TypeVar,
    Union,
    overload,
)

import svgelements
from OCP.BRepAdaptor import BRepAdaptor_Curve
from OCP.BRepBuilderAPI import BRepBuilderAPI_Transform
from OCP.Geom import Geom_BezierCurve
from OCP.GeomAbs import GeomAbs_CurveType
from OCP.GeomAdaptor import GeomAdaptor_Curve
from OCP.gp import gp_Ax1, gp_Dir, gp_Pnt, gp_Trsf, gp_Vec
from OCP.StdFail import StdFail_NotDone
from OCP.TopoDS import TopoDS, TopoDS_Edge, TopoDS_Face, TopoDS_Shape, TopoDS_Wire

from .ocp import (
    CurveOrAdaptor,
    InvalidWiresForFace,
    bezier_curve,
    circle_curve,
    curve_and_adaptor,
    curve_to_beziers,
    curve_to_polyline,
    edge_from_curve,
    edge_to_curve,
    ellipse_curve,
    faces_from_wire_soup,
    segment_curve,
    topoDS_iterator,
    wire_from_continuous_edges,
)

__all__ = [
    "import_svg_document",
    "DocumentInfo",
    "ColorAndLabel",
    "faces_from_svg_path",
    "wires_from_svg_path",
    "edges_from_svg_path",
    "continuous_edges_from_svg_path",
    "face_to_svg_path",
    "wire_to_svg_path",
    "edge_to_svg_path",
    "curve_to_svg_path",
    "SvgPathCommand",
    "format_svg",
]


flip_y = gp_Trsf()
flip_y.SetMirror(gp_Ax1(gp_Pnt(), gp_Dir(1, 0, 0)))


logger = logging.getLogger(__name__)

SvgPathCommand = Union[
    tuple[Literal["M"], float, float],
    tuple[Literal["L"], float, float],
    tuple[Literal["Q"], float, float, float, float],
    tuple[Literal["C"], float, float, float, float, float, float],
    tuple[Literal["A"], float, float, float, int, int, float, float],
    tuple[Literal["Z"]],
]


def format_svg(path: Iterable[SvgPathCommand], float_format: str = "f") -> str:
    return " ".join(
        f"{cmd[0]}{','.join(arg.__format__(float_format) for arg in cmd[1:])}"
        for cmd in path
    )


SvgPathLike = Union[str, Iterable[SvgPathCommand], svgelements.Path]
ShapeElement = svgelements.Shape
ParentElement = Union[svgelements.Group, svgelements.Use]

FaceOrWire = Union[TopoDS_Wire, TopoDS_Face]


class DocumentInfo(NamedTuple):
    width: float
    height: float


T = TypeVar("T")


class ItemsFromDocument(Iterable[T]):
    def __init__(self, elements: Iterator[T], doc_info: DocumentInfo) -> None:
        self.elements = elements
        self.doc_info = doc_info

    def __iter__(self) -> Iterator[T]:
        return self.elements


M = TypeVar("M")


@overload
def import_svg_document(
    svg_file: Union[str, pathlib.Path, TextIO],
    *,
    flip_y: bool = True,
    ignore_visibility: bool = False,
    metadata: Optional[Callable[[ShapeElement, Sequence[ParentElement]], M]],
) -> ItemsFromDocument[tuple[FaceOrWire, M]]: ...


@overload
def import_svg_document(
    svg_file: Union[str, pathlib.Path, TextIO],
    *,
    flip_y: bool = True,
    ignore_visibility: bool = False,
) -> ItemsFromDocument[FaceOrWire]: ...


def import_svg_document(
    svg_file: Union[str, pathlib.Path, TextIO],
    *,
    flip_y: bool = True,
    ignore_visibility: bool = False,
    metadata: Optional[Callable[[ShapeElement, Sequence[ParentElement]], M]] = None,
) -> Union[ItemsFromDocument[tuple[FaceOrWire, M]], ItemsFromDocument[FaceOrWire]]:
    """Import shapes from an SVG document as faces and/or wires.

    Each visible shape and path is converted to zero or more Face if it is filled,
    and to zero or more Wire if it is not filled.

    This importer does not cover the whole SVG specification,
    its most notable known limitations are:

    - degenerate and self-crossing paths may result in invalid faces and wires
    - clipping, both by clipping paths and viewport, is ignored
    - graphic properties such as line strokes and pattern fills are ignored

    Documents relying on these features need to be pre-processed externally.

    :param svg_file: input SVG document
    :param flip_y: whether to mirror the Y-coordinates
        to compensate for SVG's top left origin, defaults to True
    :param ignore_visibility: whether to ignore visibility
        attribute and process hidden elements.
    :param metadata: funtion to generate metadata from the source SVG element
    :raises IOError:
    :raises SyntaxError:
    :raises ValueError:
    """

    def doc_transform(info: DocumentInfo):
        mirror = gp_Trsf()
        mirror.SetMirror(gp_Ax1(gp_Pnt(0, info.height / 2, 0), gp_Dir(1, 0, 0)))

        def f_flip(shape: FaceOrWire) -> FaceOrWire:
            mirrored = BRepBuilderAPI_Transform(shape, mirror, False, False).Shape()
            mirrored.Reverse()
            if isinstance(shape, TopoDS_Face):
                return TopoDS.Face_s(mirrored)
            if isinstance(shape, TopoDS_Wire):
                return TopoDS.Wire_s(mirrored)
            else:
                raise AssertionError(f"somehow got unexpected shape {shape}")

        def f_identity(shape: FaceOrWire) -> FaceOrWire:
            return shape

        return f_flip if flip_y else f_identity

    def process_wire(
        wires: list[TopoDS_Wire],
        is_filled: bool,
    ) -> Iterator[FaceOrWire]:
        if is_filled:
            try:
                yield from faces_from_wire_soup(wires)
            except InvalidWiresForFace:
                logger.warning("filled shape could not be converted to face")  # TODO
                yield from wires
        else:
            yield from wires

    if metadata:
        wires_from_doc = wires_from_svg_document(
            svg_file,
            ignore_visibility=ignore_visibility,
            metadata_factory=metadata,
        )
        transform = doc_transform(wires_from_doc.doc_info)
        items = (
            (transform(face_or_wire), m)
            for wires, is_filled, m in wires_from_doc
            for face_or_wire in process_wire(wires, is_filled)
        )
        return ItemsFromDocument(items, wires_from_doc.doc_info)
    else:
        wires_from_doc = wires_from_svg_document(
            svg_file,
            ignore_visibility=ignore_visibility,
            metadata_factory=None,
        )
        transform = doc_transform(wires_from_doc.doc_info)
        items = (
            transform(face_or_wire)
            for wires, is_filled, _metadata in wires_from_doc
            for face_or_wire in process_wire(wires, is_filled)
        )
        return ItemsFromDocument(items, wires_from_doc.doc_info)


class ColorAndLabel:
    def __init__(
        self,
        element: ShapeElement,
        parents: Sequence[ParentElement],
        label_by: str = "id",
    ) -> None:
        self.fill_color = self._color(element.fill)
        self.stroke_color = self._color(element.stroke) or (0, 0, 0, 1)
        self.label = self._label(element, label_by)
        self.parent_labels = tuple(self._label(parent, label_by) for parent in parents)

    def color_for(self, shape: TopoDS_Shape):
        """Fill color if shape should be filled stroke color otherwise."""
        filled = not isinstance(shape, (TopoDS_Wire, TopoDS_Edge))
        return self.fill_color if filled and self.fill_color else self.stroke_color

    @staticmethod
    def _color(
        color: Union[svgelements.Color, None]
    ) -> Union[tuple[float, float, float, float], None]:
        if color and color.value:  # type: ignore
            try:
                rgba = color.red, color.green, color.blue, color.alpha  # type: ignore
                return tuple(float(v) / 255 for v in rgba)  # type: ignore
            except TypeError:
                return 0, 0, 0, 1

    @staticmethod
    def _label(element: Union[ShapeElement, ParentElement], label_by: str):
        try:
            return str(element.values[label_by])  # type: ignore
        except (KeyError, AttributeError):
            return ""

    @classmethod
    def Label_by(cls, label_by: str = "id"):
        def f(element: ShapeElement, parents: Sequence[ParentElement]):
            return cls(element, parents, label_by=label_by)

        return f


####


def faces_from_svg_path(path: SvgPathLike) -> Iterable[TopoDS_Face]:
    """Create faces from an SVG path.

    :param SvgPathLike path: input SVG path
    :yield: faces
    :raises SyntaxError:
    :raises ValueError:
    """
    return faces_from_wire_soup(wires_from_svg_path(path))


def wires_from_svg_element(element: ShapeElement) -> Iterable[TopoDS_Wire]:
    def check_unskewed_transform() -> Union[tuple[float, float, float], None]:
        o = svgelements.Point(0, 0) * element.transform
        x = svgelements.Point(1, 0) * element.transform
        y = svgelements.Point(0, 1) * element.transform
        is_skewed = abs(o.angle_to(y) - o.angle_to(x)) != pi / 2  # type: ignore
        if not is_skewed:
            return o.distance_to(x), o.distance_to(y), o.angle_to(x)  # type: ignore

    element.reify()

    if isinstance(element, (svgelements.Circle, svgelements.Ellipse)) and (
        scale_and_angle := check_unskewed_transform()
    ):
        cx = float(element.cx)  # type: ignore
        cy = float(element.cy)  # type: ignore
        origin = svgelements.Point(cx, cy) * element.transform
        center = gp_Pnt(origin.real, origin.imag, 0)  # type: ignore
        scale_x, scale_y, angle = scale_and_angle
        r1 = float(element.rx) * scale_x  # type: ignore
        r2 = float(element.ry) * scale_y  # type: ignore
        if r1 == r2:
            curve = circle_curve(r1, center=center)
        else:
            curve = ellipse_curve(r1, r2, center=center, rotation=degrees(angle))
        yield wire_from_continuous_edges((edge_from_curve(curve),))

    elif path := svg_element_to_path(element):
        yield from wires_from_svg_path(path)


def svg_element_to_path(element: ShapeElement):
    path = svgelements.Path(element)
    if len(path):
        path.reify()
        return path


def wires_from_svg_path(path: SvgPathLike) -> Iterable[TopoDS_Wire]:
    """Create wires from an SVG path.

    :param SvgPathLike path: input SVG path
    :yield: wires
    :raises SyntaxError:
    :raises ValueError:
    """

    for edges, closed in continuous_edges_from_svg_path(path):
        yield wire_from_continuous_edges(edges, closed=closed)


def edges_from_svg_path(path: SvgPathLike) -> Iterable[TopoDS_Edge]:
    """Create edges from an SVG path.

    :param SvgPathLike path: input SVG path
    :yield: edges
    :raises SyntaxError:
    :raises ValueError:
    """
    for edges, _ in continuous_edges_from_svg_path(path):
        yield from edges


def continuous_edges_from_svg_path(
    path: SvgPathLike,
) -> Iterable[tuple[Iterable[TopoDS_Edge], bool]]:
    def p(c: Union[svgelements.Point, None]):
        if c is None:  # pragma: nocover
            logger.warning("found None point in %s", path)
            return gp_Pnt(0, 0, 0)
        else:
            return gp_Pnt(c.real, c.imag, 0)  # type: ignore

    def curves_from_segment(
        segment: Union[
            svgelements.Line,
            svgelements.QuadraticBezier,
            svgelements.CubicBezier,
            svgelements.Arc,
        ]
    ):
        if isinstance(segment, svgelements.Move):
            pass
        elif isinstance(segment, svgelements.Line):
            if segment.start != segment.end:
                return segment_curve(p(segment.start), p(segment.end))
        elif isinstance(segment, svgelements.QuadraticBezier):
            return bezier_curve(p(segment.start), p(segment.control), p(segment.end))
        elif isinstance(segment, svgelements.CubicBezier):
            return bezier_curve(
                p(segment.start),
                p(segment.control1),
                p(segment.control2),
                p(segment.end),
            )
        elif isinstance(segment, svgelements.Arc):
            start_angle = segment.theta
            end_angle = segment.theta + segment.delta
            return ellipse_curve(
                segment.radius.real,  # type: ignore
                segment.radius.imag,  # type: ignore
                start_angle=min(start_angle, end_angle),
                end_angle=max(end_angle, start_angle),
                clockwise=end_angle > start_angle,
                center=p(segment.center),
                rotation=degrees(segment.get_rotation()),
            )
        elif isinstance(segment, svgelements.Close):
            if segment.start != segment.end:
                return segment_curve(p(segment.start), p(segment.end))
        else:  # pragma: nocover
            logger.warning(f"unexpected segment type: {type(segment)}")
            return segment_curve(p(segment.start), p(segment.end))

    def edges_from_path(
        path: svgelements.Path,
    ):
        for segment in path:  # type: ignore
            try:
                if curve := curves_from_segment(segment):  # type: ignore
                    yield edge_from_curve(curve)
            except (StdFail_NotDone, ValueError):  # pragma: nocover
                logger.debug("invalid %s", _SegmentInPath(segment, path))

    path = _path_from_SvgPathLike(path)
    for subpath, closed in _continuous_subpaths(path):
        if edges := list(edges_from_path(subpath)):
            yield edges, closed


class _SegmentInPath:
    def __init__(self, segment: Any, path: svgelements.Path) -> None:
        self.segment = segment
        self.path = path

    def __str__(self) -> str:
        d = self.path.d()  # type: ignore
        return f"`{self.segment}` segment in path `{d}`"


####


def face_to_svg_path(
    face: TopoDS_Face,
    *,
    tolerance: float,
    use_cubics: bool = True,
    use_quadratics: bool = True,
    use_arcs: bool = True,
    split_full_arcs: bool = True,
    with_first_move: bool = True,
) -> Iterable[SvgPathCommand]:
    return chain.from_iterable(
        wire_to_svg_path(
            TopoDS.Wire_s(wire),
            tolerance=tolerance,
            use_cubics=use_cubics,
            use_quadratics=use_quadratics,
            use_arcs=use_arcs,
            split_full_arcs=split_full_arcs,
            with_first_move=with_first_move and i == 0,
        )
        for i, wire in enumerate(topoDS_iterator(face))
    )


def wire_to_svg_path(
    wire: TopoDS_Wire,
    *,
    tolerance: float,
    use_cubics: bool = True,
    use_quadratics: bool = True,
    use_arcs: bool = True,
    split_full_arcs: bool = True,
    with_first_move: bool = True,
) -> Iterable[SvgPathCommand]:
    return chain.from_iterable(
        edge_to_svg_path(
            TopoDS.Edge_s(edge),
            tolerance=tolerance,
            use_cubics=use_cubics,
            use_quadratics=use_quadratics,
            use_arcs=use_arcs,
            split_full_arcs=split_full_arcs,
            with_first_move=with_first_move and i == 0,
        )
        for i, edge in enumerate(topoDS_iterator(wire))
    )


def edge_to_svg_path(
    edge: TopoDS_Edge,
    *,
    tolerance: float,
    use_cubics: bool = True,
    use_quadratics: bool = True,
    use_arcs: bool = True,
    split_full_arcs: bool = True,
    with_first_move: bool = True,
) -> Iterable[SvgPathCommand]:
    return curve_to_svg_path(
        edge_to_curve(edge),
        tolerance=tolerance,
        use_cubics=use_cubics,
        use_quadratics=use_quadratics,
        use_arcs=use_arcs,
        split_full_arcs=split_full_arcs,
        with_first_move=with_first_move,
    )


def curve_to_svg_path(
    curve_or_adaptor: CurveOrAdaptor,
    *,
    tolerance: float,
    use_cubics: bool = True,
    use_quadratics: bool = True,
    use_arcs: bool = True,
    split_full_arcs: bool = True,
    with_first_move: bool = True,
) -> Iterable[SvgPathCommand]:
    _curve, adaptor = curve_and_adaptor(curve_or_adaptor)
    curve_type = adaptor.GetType()

    try:
        if curve_type == GeomAbs_CurveType.GeomAbs_Line:
            p0 = adaptor.Value(adaptor.FirstParameter())
            p1 = adaptor.Value(adaptor.LastParameter())
            if with_first_move:
                yield "M", p0.X(), p0.Y()
            yield "L", p1.X(), p1.Y()

        elif use_arcs and curve_type in (
            GeomAbs_CurveType.GeomAbs_Circle,
            GeomAbs_CurveType.GeomAbs_Ellipse,
        ):
            if curve_type == GeomAbs_CurveType.GeomAbs_Circle:
                circle = adaptor.Circle()
                r1 = r2 = circle.Radius()
                axis = circle.XAxis()
            else:
                ellipse = adaptor.Ellipse()
                r1, r2 = ellipse.MajorRadius(), ellipse.MinorRadius()
                axis = ellipse.XAxis()

            yield from ellipse_to_svg_path(
                adaptor,
                r1,
                r2,
                axis,
                split_full_arcs=split_full_arcs,
                with_first_move=with_first_move,
            )

        elif use_cubics or use_quadratics:
            beziers = curve_to_beziers(
                adaptor, tolerance=tolerance, max_degree=3 if use_cubics else 2
            )
            for i, bezier in enumerate(beziers):
                yield from bezier_to_svg_path(
                    bezier,
                    use_quadratics=use_quadratics,
                    with_first_move=with_first_move and i == 0,
                )
        else:
            yield from polyline_to_svg_path(
                curve_to_polyline(adaptor, tolerance=tolerance),
                with_first_move=with_first_move,
                closed=adaptor.IsClosed(),
            )
    except Exception as e:  # pragma: nocover
        logger.error(
            "failed to convert %s curve, falling back to polyline approximation",
            curve_type,
            exc_info=e,
        )
        yield from polyline_to_svg_path(
            curve_to_polyline(adaptor, tolerance=tolerance),
            with_first_move=with_first_move,
            closed=adaptor.IsClosed(),
        )


def ellipse_to_svg_path(
    adaptor_curve: Union[GeomAdaptor_Curve, BRepAdaptor_Curve],
    r1: float,
    r2: float,
    axis: gp_Ax1,
    with_first_move: bool = True,
    split_full_arcs: bool = True,
) -> Iterable[SvgPathCommand]:
    t0 = adaptor_curve.FirstParameter()
    t1 = adaptor_curve.LastParameter()
    p0 = adaptor_curve.Value(t0)
    pm = adaptor_curve.Value((t1 + t0) / 2.0)
    p1 = adaptor_curve.Value(t1)

    a = gp_Vec(0, 0, 1).DotCross(gp_Vec(pm, p0), gp_Vec(pm, p1))

    angle = axis.Direction().AngleWithRef(gp_Dir(1, 0, 0), gp_Dir(0, 0, -1))
    large_arc = 1 if t1 - t0 > pi else 0
    sweep = 1 if a < 0 else 0

    if with_first_move:
        yield "M", p0.X(), p0.Y()

    if adaptor_curve.IsClosed():
        if split_full_arcs:
            yield "A", r1, r2, 180, 1, 1, pm.X(), pm.Y()
            yield "A", r1, r2, 180, 1, 1, p1.X(), p1.Y()
        else:
            almost_one = 1 - 1e-8  # seem to be a close as we can get for most viewers
            pq = adaptor_curve.Value(t0 + (t1 - t0) * almost_one)
            yield "A", r1, r2, 360 * almost_one, 1, 0, pq.X(), pq.Y()
    else:
        yield "A", r1, r2, degrees(angle), large_arc, sweep, p1.X(), p1.Y()


def bezier_to_svg_path(
    bezier: Geom_BezierCurve,
    *,
    use_quadratics: bool = True,
    with_first_move: bool = True,
) -> Iterable[SvgPathCommand]:
    degree = bezier.Degree()

    p0 = bezier.Pole(1)
    if with_first_move:
        yield "M", p0.X(), p0.Y()

    if degree == 3:
        p1 = bezier.Pole(2)
        p2 = bezier.Pole(3)
        p3 = bezier.Pole(4)
        yield "C", p1.X(), p1.Y(), p2.X(), p2.Y(), p3.X(), p3.Y()

    elif degree == 2:
        p1 = bezier.Pole(2)
        p2 = bezier.Pole(3)
        if use_quadratics:
            yield "Q", p1.X(), p1.Y(), p2.X(), p2.Y()
        else:
            p1b = gp_Vec(p0.X(), p0.Y(), 0).Added(gp_Vec(p0, p1).Multiplied(2 / 3))
            p2b = gp_Vec(p2.X(), p2.Y(), 0).Added(gp_Vec(p2, p1).Multiplied(2 / 3))
            yield "C", p1b.X(), p1b.Y(), p2b.X(), p2b.Y(), p2.X(), p2.Y()

    elif degree == 1:
        p1 = bezier.Pole(2)
        yield "L", p1.X(), p1.Y()

    else:
        raise ValueError(f"could not convert {bezier} to SVG path")


def polyline_to_svg_path(
    points: Iterable[gp_Pnt],
    *,
    with_first_move: bool = True,
    closed: bool = False,
) -> Iterable[SvgPathCommand]:
    points = list(points)
    if points:
        first, *others = points

        if with_first_move:
            yield "M", first.X(), first.Y()
        for point in others:
            yield "L", point.X(), point.Y()

        if closed:
            yield "Z",


####


@overload
def wires_from_svg_document(
    svg_file: Union[str, pathlib.Path, TextIO],
    metadata_factory: Callable[[ShapeElement, Sequence[ParentElement]], M],
    *,
    ignore_visibility: bool = False,
) -> ItemsFromDocument[tuple[list[TopoDS_Wire], bool, M]]: ...


@overload
def wires_from_svg_document(
    svg_file: Union[str, pathlib.Path, TextIO],
    metadata_factory: None,
    *,
    ignore_visibility: bool = False,
) -> ItemsFromDocument[tuple[list[TopoDS_Wire], bool, None]]: ...


def wires_from_svg_document(
    svg_file: Union[str, pathlib.Path, TextIO],
    metadata_factory: Optional[Callable[[ShapeElement, Sequence[ParentElement]], M]],
    *,
    ignore_visibility: bool = False,
) -> Union[
    ItemsFromDocument[tuple[list[TopoDS_Wire], bool, M]],
    ItemsFromDocument[tuple[list[TopoDS_Wire], bool, None]],
]:
    elements = find_shapes_svg_in_document(
        svg_file, ignore_visibility=ignore_visibility
    )

    def is_filled(element: ShapeElement):
        fill = element.fill
        return fill.value is not None  # type: ignore

    if callable(metadata_factory):
        wires = (
            (
                list(wires_from_svg_element(source_element)),
                is_filled(source_element),
                metadata_factory(source_element, source_parents),
            )
            for source_element, source_parents in elements
        )
        return ItemsFromDocument(wires, elements.doc_info)
    else:
        wires = (
            (
                list(wires_from_svg_element(source_element)),
                is_filled(source_element),
                None,
            )
            for source_element, _source_parents in elements
        )
        return ItemsFromDocument(wires, elements.doc_info)


def find_shapes_svg_in_document(
    svg_file: Union[str, pathlib.Path, TextIO],
    *,
    ignore_visibility: bool = False,
) -> ItemsFromDocument[tuple[ShapeElement, tuple[ParentElement, ...]]]:
    def walk_svg_element(
        element: svgelements.SVGElement, parents: tuple[ParentElement, ...] = ()
    ) -> Iterator[tuple[ShapeElement, tuple[ParentElement, ...]]]:
        if isinstance(element, ShapeElement):
            yield element, parents
        elif isinstance(element, (svgelements.Group, svgelements.Use)):
            new_parents = *parents, element
            for child in element:  # type: ignore
                yield from walk_svg_element(child, new_parents)  # type: ignore

    parsed_svg = svgelements.SVG.parse(  # type: ignore
        (
            resolve_path(svg_file)
            if isinstance(svg_file, (str, pathlib.Path))
            else svg_file
        ),
        parse_display_none=ignore_visibility,
        ppi=25.4,  # inches to millimiters
    )

    def elements():
        for element, parents in walk_svg_element(parsed_svg):  # type: ignore
            if not ignore_visibility:
                try:
                    visibility = element.values["visibility"]  # type: ignore
                    if visibility in ("hidden", "collapse"):  # type: ignore
                        continue
                except (KeyError, AttributeError):
                    pass

            if isinstance(element, svgelements.Path):
                if len(element):
                    yield element, parents

            elif isinstance(element, svgelements.Shape):
                path = svgelements.Path(element)
                if len(path):
                    path.reify()
                    yield element, parents

    doc_info = DocumentInfo(parsed_svg.width, parsed_svg.height)  # type: ignore
    return ItemsFromDocument(elements(), doc_info)


def resolve_path(path: Union[pathlib.Path, str]) -> str:
    if not isinstance(path, pathlib.Path):
        path = pathlib.Path(path)
    return str(path.expanduser().resolve())


def _path_from_SvgPathLike(path: SvgPathLike) -> svgelements.Path:
    if isinstance(path, svgelements.Path):
        return path

    if not isinstance(path, str):
        path = format_svg(path)

    try:
        return svgelements.Path(str(path))
    except Exception:
        # TODO proper syntax error, would need to come from within svgpathelements
        raise ValueError(f"could not make svg path from: {path!r}")


def _continuous_subpaths(
    path: svgelements.Path,
) -> Iterator[tuple[svgelements.Path, bool]]:
    for subpath in path.as_subpaths():
        if subpath:
            is_closed = bool(subpath[0] == subpath[-1])  # type: ignore
            yield svgelements.Path(subpath), is_closed



import logging
import tempfile
import time
from collections import Counter
from contextlib import contextmanager
from pathlib import Path
from urllib.request import urlretrieve

from ocpsvg import ColorAndLabel, import_svg_document


@contextmanager
def timer(title: str):
    t0 = time.perf_counter()
    try:
        yield
    except Exception:
        raise
    finally:
        t1 = time.perf_counter()
        logging.info(f"{title}: {t1-t0:0.3f}s")


logging.basicConfig(level=logging.DEBUG)


TIGER_URL = "https://upload.wikimedia.org/wikipedia/commons/f/fd/Ghostscript_Tiger.svg"
tiger_path = Path(tempfile.gettempdir()) / "Ghostscript_Tiger.svg"

if not tiger_path.is_file():
    urlretrieve(TIGER_URL, tiger_path)
if not tiger_path.is_file():
    raise IOError(f"could not retrieve {TIGER_URL}")


with timer(f"import {tiger_path}"):
    imported = list(import_svg_document(tiger_path, metadata=ColorAndLabel))

logging.info(
    " ".join(
        f"{v}*{k.__qualname__}"
        for k, v in Counter(map(type, (o for o, _ in imported))).items()
    )
)

show_object = globals().get("show_object")
if callable(show_object):
    for face_or_wire, color_and_label in imported:
        show_object(
            face_or_wire,
            f"{color_and_label.label} #{id(face_or_wire):x}",
            dict(color=color_and_label.color),
        )



from OCP.BRepGProp import BRepGProp, BRepGProp_Face
from OCP.BRepTools import BRepTools
from OCP.gp import gp_Pnt, gp_Vec
from OCP.GProp import GProp_GProps
from OCP.TopoDS import TopoDS_Face


def face_area(face: TopoDS_Face):
    properties = GProp_GProps()
    BRepGProp.SurfaceProperties_s(face, properties)

    return properties.Mass()


def face_normal(face: TopoDS_Face) -> gp_Vec:
    u0, u1, v0, v1 = BRepTools.UVBounds_s(face)
    return face_normal_at_uv(face, (u0 + u1) / 2, (v0 + v1) / 2)


def face_normal_at_uv(face: TopoDS_Face, u: float, v: float) -> gp_Vec:
    gp_pnt = gp_Pnt()
    normal = gp_Vec()
    BRepGProp_Face(face).Normal(u, v, gp_pnt, normal)
    return normal



from .svg import (
    ColorAndLabel,
    DocumentInfo,
    ParentElement,
    ShapeElement,
    SvgPathCommand,
    curve_to_svg_path,
    edge_to_svg_path,
    edges_from_svg_path,
    face_to_svg_path,
    faces_from_svg_path,
    format_svg,
    import_svg_document,
    wire_to_svg_path,
    wires_from_svg_path,
)



import os
import re
from dataclasses import dataclass
from io import StringIO
from itertools import cycle
from math import pi
from tempfile import NamedTemporaryFile
from typing import Any, Sequence, Union

import pytest
import svgelements
from OCP.Geom import Geom_Circle, Geom_Curve, Geom_Ellipse
from OCP.GeomAbs import GeomAbs_CurveType
from OCP.gp import gp_Vec
from OCP.TopoDS import TopoDS, TopoDS_Edge, TopoDS_Face, TopoDS_Shape, TopoDS_Wire
from pytest import approx, raises

from ocpsvg.ocp import (
    bezier_curve,
    bounding_box,
    circle_curve,
    edge_from_curve,
    edge_to_curve,
    ellipse_curve,
    face_inner_wires,
    is_wire_closed,
    segment_curve,
    topoDS_iterator,
    wire_from_continuous_edges,
)
from ocpsvg.svg import (
    ColorAndLabel,
    SvgPathCommand,
    _SegmentInPath,
    bezier_to_svg_path,
    edge_to_svg_path,
    edges_from_svg_path,
    faces_from_svg_path,
    find_shapes_svg_in_document,
    format_svg,
    import_svg_document,
    polyline_to_svg_path,
    svg_element_to_path,
    wire_to_svg_path,
    wires_from_svg_path,
)

from .ocp import face_area, face_normal
from .test_ocp import XY, Pnt, as_Pnt, as_Pnts, as_tuple


class SvgPath(list[SvgPathCommand]):
    def __str__(self) -> str:
        return format_svg(self)


@pytest.mark.parametrize(
    "points, closed, svg_d",
    [
        ([(0, 1), (2, 3)], False, "M0,1 L 2,3"),
        ([(0, 1), (2, 3), (2, 5)], False, "M0,1 L 2,3 L 2,5"),
        ([(0, 1), (2, 3), (2, 5), (5, 5)], False, "M0,1 L 2,3 L 2,5 L 5,5"),
        ([(0, 1), (2, 3), (2, 5), (5, 5)], True, "M0,1 L 2,3 L 2,5 L 5,5 Z"),
    ],
)
def test_polyline_to_svg(points: Sequence[XY], closed: bool, svg_d: str):
    path = SvgPath(polyline_to_svg_path(map(as_Pnt, points), closed=closed))
    assert svg_path_tokens(path) == approx(svg_path_tokens(svg_d), abs=1e-4), str(path)


@pytest.mark.parametrize(
    "points, svg_d, opts",
    [
        ([(0, 1), (2, 3)], "M0,1 L 2,3", {}),
        ([(0, 1), (2, 3), (2, 5)], "M0,1 Q 2,3 2,5", {}),
        ([(0, 1), (2, 3), (2, 5), (5, 5)], "M0,1 C 2,3 2,5 5,5", {}),
    ],
)
def test_bezier_to_svg(points: Sequence[XY], svg_d: str, opts: dict[str, Any]):
    path = SvgPath(bezier_to_svg_path(bezier_curve(*as_Pnts(*points)), **opts))
    assert svg_path_tokens(path) == approx(svg_path_tokens(svg_d), abs=1e-4), str(path)


def test_bezier_to_svg_error():
    bez_curve = bezier_curve(*as_Pnts((0, 1), (2, 3), (2, 5), (5, 5), (10, 5)))
    with pytest.raises(ValueError):
        SvgPath(bezier_to_svg_path(bez_curve))


@pytest.mark.parametrize(
    "curve, svg_d, opts",
    [
        (segment_curve(*as_Pnts((0, 1), (2, 3))), "M0,1 L 2,3", {}),
        (bezier_curve(*as_Pnts((0, 1), (3, 2), (4, 5))), "M 0,1 Q 3,2,4,5", {}),
        (
            bezier_curve(*as_Pnts((0, 1), (3, 2), (4, 5))),
            "M 0.0,1.0 C 2.0,1.666666,3.333333,3.0,4.0,5.0",
            dict(use_quadratics=False),
        ),
        (
            bezier_curve(*as_Pnts((0, 1), (3, 2), (4, 5), (7, 8))),
            "M 0,1 C 3,2,4,5,7,8",
            {},
        ),
        (circle_curve(2), "M2,0 A2,2,180,1,1,-2,0 A2,2,180,1,1,2,0", {}),
        (circle_curve(2), "M2,0 A2,2,359.9999,1,0,2,-0", dict(split_full_arcs=False)),
        (circle_curve(2, 90, 180), "M-2.0,0 A2.0,2.0,0.0,0,0,0,2.0", {}),
        (ellipse_curve(3, 1), "M3,0 A3,1,360,1,0,3,0", dict(split_full_arcs=False)),
        (ellipse_curve(3, 1), "M3,0 A3,1,180,1,1,-3,0 A3,1,180,1,1,3,0", {}),
        (ellipse_curve(3, 1, 90, 180), "M-3.0,0 A3.0,1.0,0.0,0,0,0,1.0", {}),
    ],
)
def test_edge_to_svg(curve: Geom_Curve, svg_d: str, opts: dict[str, Any]):
    edge = edge_from_curve(curve)
    path = SvgPath(edge_to_svg_path(edge, tolerance=1e-5, **opts))
    assert svg_path_tokens(path) == approx(svg_path_tokens(svg_d), abs=1e-4), str(path)


@pytest.mark.parametrize(
    "curve",
    [
        segment_curve(*as_Pnts((0, 1), (2, 3))),
        bezier_curve(*as_Pnts((0, 1), (3, 2), (4, 5))),
        bezier_curve(*as_Pnts((0, 1), (3, 2), (4, 5), (7, 8))),
        circle_curve(2),
        circle_curve(2, 90, 180),
        ellipse_curve(3, 1, 90, 180),
    ],
)
def test_polyline_apprx(curve: Geom_Curve):
    edge = edge_from_curve(curve)
    path = SvgPath(
        edge_to_svg_path(
            edge, tolerance=1e-5, use_arcs=False, use_cubics=False, use_quadratics=False
        )
    )
    assert set(seg[0] for seg in path).issubset(set("MLZ"))


@pytest.mark.parametrize(
    "curves, svg_d, opts",
    [
        (
            [
                segment_curve(Pnt(-1, 4), Pnt(0, 1)),
                bezier_curve(Pnt(0, 1), Pnt(3, 2), Pnt(4, 5)),
            ],
            "M-1,4 L 0,1" "Q 3,2,4,5",
            {},
        ),
        (
            [
                segment_curve(Pnt(-1, 4), Pnt(0, 1)),
                bezier_curve(Pnt(0, 1), Pnt(3, 2), Pnt(4, 5)),
            ],
            "M-1,4 L 0,1" "C 2.0,1.666666,3.333333,3.0,4.0,5.0",
            dict(use_quadratics=False),
        ),
    ],
)
def test_wire_to_svg(curves: list[Geom_Curve], svg_d: str, opts: dict[str, Any]):
    wire = wire_from_continuous_edges(map(edge_from_curve, curves))
    path = SvgPath(wire_to_svg_path(wire, tolerance=1e-5, **opts))
    assert svg_path_tokens(path) == approx(svg_path_tokens(svg_d), abs=1e-4), str(path)


def test_arc_flags():
    c = pi * 45**2
    s = (45 * 2) ** 2
    cases = [
        ("M  80  80 A 45 45, 0, 0, 0, 125 125 L 125  80 Z", 1 / 4 * c),
        ("M 230  80 A 45 45, 0, 1, 0, 275 125 L 275  80 Z", 3 / 4 * c + 1 / 4 * s),
        ("M  80 230 A 45 45, 0, 0, 1, 125 275 L 125 230 Z", 1 / 4 * (s - c)),
        ("M 230 230 A 45 45, 0, 1, 1, 275 275 L 275 230 Z", 3 / 4 * c),
    ]
    for path, area in cases:
        res = list(faces_from_svg_path(path))
        assert len(res) == 1
        assert isinstance(res[0], TopoDS_Face)
        assert face_area(res[0]) == approx(area)


def test_path_regression1():
    """this path used to generate segments below tolerance when being closed"""
    d = """
    M 33.5373470102,42.5176389595
    A 4.09636,4.09636 0.0 0,1 29.4409888981,46.6139970715
    A 4.09636,4.09636 0.0 0,1 25.3446307861,42.5176389595
    A 4.09636,4.09636 0.0 0,1 29.4409888981,38.4212808475
    A 4.09636,4.09636 0.0 0,1 33.5373470102,42.5176389595
    Z
    """
    wires = list(wires_from_svg_path(d))
    assert len(wires) == 1
    assert is_wire_closed(wires[0])


def test_path_regression2():
    """the small arc in this path used to no be handled right"""
    d = """
    m 11.76429,44.366052 
    c 0.865598,-0.538449 2.987334,-1.654154 5.219562,-3.0173931.860913,-1.136474
    4.482602,-2.785758 7.441652,-4.045502 2.033322,-0.865638 4.174966,-1.53082
    6.465079,-1.859277 2.328204,-0.333919 4.652868,-0.296857 7.012161,0.09842 
    0.0066,0.0016 0.01317,0.0032 0.01975,0.0049 1.236231,0.306435 2.433142,0.808671
    3.617195,1.509105 1.146272,0.678084 2.25166,1.524577 3.349647,2.518228 
    2.211881,2.001698 4.261035,4.482385 6.312848,6.896273 2.52664,2.972508 
    4.591476,5.255564 6.895181,7.028956 0.52806,0.406501 1.03584,0.75893 
    1.522741,1.061625 0.901859,0.560664 1.808731,0.981731 2.68984,1.286418 
    2.090466,0.722882 3.751425,0.698313 4.665988,0.720769 -0.895985,-0.382042 
    -2.163264,-1.190245 -3.678877,-2.48684 -0.641461,-0.548765 -1.26771,-1.128549 
    -1.959619,-1.780292 -0.371036,-0.349498 -0.763728,-0.727866 -1.188041,-1.152766 
    -2.037319,-2.040141 -3.506627,-3.812847 -6.177306,-7.011402 -1.988611,-2.381671 
    -4.248901,-5.050679 -6.738801,-7.211117 -1.257303,-1.090937 -2.575615,-2.057722 
    -3.984539,-2.846659 -1.462016,-0.818666 -2.971376,-1.417477 -4.550841,-1.782702 
    -0.02614,-0.006 -0.05233,-0.01203 -0.07858,-0.01796
    a 0.38591405,0.03021177 15.61296 0 0 -0.113793,-0.0224
    c -2.725017,-0.415945 -5.42773,-0.409008 -8.116554,0.06309 -2.609069,0.45809 
    -4.995593,1.317744 -7.161968,2.400596 -3.147128,1.573078 -5.751218,3.587805 
    -7.400338,5.091117 -1.968561,1.794511 -3.433935,3.806292 -4.062386,4.554845
    z
    """
    wires = list(wires_from_svg_path(d))
    assert len(wires) == 1
    assert is_wire_closed(wires[0])


def test_path_regression3():
    """This path would not produce a face with 3 holes
    when the wires were not re-oriented correctly"""
    d = """
    M 62.86,35.92
    C 78.51,36.69 76.47,63.17 59.62,59.32 42.78,55.47 47.20,35.16 62.86,35.92
    Z
    m -3.99,12.40
    c -6.56,-15.20 -10.92,10.45 0.01,0.04
    M 67.41,55
    c -1.93,-16.98 -16.51,6.21 0,0
    z
    m 0.35,-8.49
    c 1.41,-17.03 -17.77,3.09 0.03,0.05
    """
    faces = list(faces_from_svg_path(d))
    assert len(faces) == 1
    assert len(face_inner_wires(faces[0])) == 3


def test_arcs_path_to_wire():
    """this continuous path introduces small discontinuities when making the edges"""
    res = list(
        wires_from_svg_path(
            "M 10 315 L 110 215"
            "A 30 50 0 0 1 162.55 162.45"
            "L 172.55 152.45"
            "A 30 50 -45 0 1 215.1 109.9"
            "L 315 10"
        )
    )
    assert len(res) == 1
    assert isinstance(res[0], TopoDS_Wire)


def test_path_from_segments():
    res = list(wires_from_svg_path([("M", 0.0, 0.0), ("L", 1.0, 2.0)]))
    assert len(res) == 1
    assert isinstance(res[0], TopoDS_Wire)


@pytest.mark.parametrize(
    "svg_d",
    [
        "M 10 5 L 15 5 L 15 5 L 15 10",
        "M 0,0 L 1,1 Q 1,1 1,1 L 2,3",
    ],
)
def test_empty_segments(svg_d: str):
    res = list(wires_from_svg_path(svg_d))
    assert len(res) == 1
    assert isinstance(res[0], TopoDS_Wire)


def test_empty_paths():
    assert not list(edges_from_svg_path(""))
    assert not list(wires_from_svg_path(""))
    assert not list(faces_from_svg_path(""))


@pytest.mark.parametrize(
    "svg_d",
    [
        "M 0,1,2",
        pytest.param(
            "M 0,0 N 1,2", marks=pytest.mark.xfail(reason="svgelements accepts it")
        ),
    ],
)
def test_invalid_path(svg_d: str):
    with raises(ValueError):
        list(edges_from_svg_path(svg_d))


@pytest.mark.parametrize(
    "svg_d, expected_count",
    [
        ("M 0,0", 0),
        ("M 0,0 v 1 h 1 z", 1),
        ("M 0,0 v 1 h 1", 1),
        ("M 10 80 Q 95 10 180 80", 1),
        ("M 10 80 C 40 10, 65 10, 95 80 S 150 150, 180 80", 1),
    ],
)
def test_faces_from_svg_path(svg_d: str, expected_count: int):
    res = list(faces_from_svg_path(svg_d))
    assert len(res) == expected_count
    assert all(isinstance(x, TopoDS_Face) for x in res)


@pytest.mark.parametrize(
    "svg_d, expected_count",
    [
        ("M 0,0 v 1 h 1", 2),
        ("M 0,0 v 1 M 1,0 v 2", 1 + 1),
        ("M 0,0 v 1 h 1 z M 2,0 v 1 h 1", 3 + 2),
        ("M 0,0 v 1 M 1,0 v 2", 2),
    ],
)
def test_edges_from_svg_path(svg_d: str, expected_count: int):
    res = list(edges_from_svg_path(svg_d))
    assert len(res) == expected_count
    assert all(isinstance(x, TopoDS_Edge) for x in res)


@pytest.mark.parametrize(
    "svg_d, expected_count",
    [
        ("M 0,0 v 1 h 1", 1),
        ("M 0,0 v 1 M 1,0 v 2", 2),
        ("M 0,0 v 1 h 1 z M 2,0 v 1 h 1", 2),
        ("M 0,0 v 1 M 1,0 v 2", 2),
    ],
)
def test_wires_from_svg_path(svg_d: str, expected_count: int):
    res = list(wires_from_svg_path(svg_d))
    assert len(res) == expected_count
    assert all(isinstance(x, TopoDS_Wire) for x in res)


@pytest.mark.parametrize("count", range(3, 10))
def test_concentric_path_nesting(count: int):
    n = count // 2
    if count % 2:
        expected_hole_counts = [0] + [1] * n
    else:
        expected_hole_counts = [1] * n
    res = list(faces_from_svg_path(nested_squares_path(count)))
    assert hole_counts(res) == expected_hole_counts


def test_path_nesting():
    """This path:
    ```
        .-----------------------------------.
        |   .----------------------------.  |
        |   | .--------------.  .-----.  |  |
        |   | | .---.        |  `-----'  |  |
        |   | | |   |  .---. |  .-----.  |  |
        |   | | `---'  `---' |  |     |  |  |
        |   | `--------------'  `-----'  |  |
        |   `----------------------------'  |
        `-----------------------------------'
    ```
    should translate to these faces (order irrelevant):
    ```
        AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
        AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
        AAAAA BBBBBBBBBBBBBBBB  CCCCCCC  AAAA
        AAAAA BBB   BBBBBBBBBB  CCCCCCC  AAAA
        AAAAA BBB   BBB     BB  DDDDDDD  AAAA
        AAAAA BBB   BBB     BB  DDDDDDD  AAAA
        AAAAA BBBBBBBBBBBBBBBB  DDDDDDD  AAAA
        AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
        AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    ```
    """
    res = list(
        faces_from_svg_path(
            "M 1,4 L 16,4 L 16,13 L 1,13 L 1,4 Z"
            "M 2,5 L 15,5 L 15,12 L 2,12 L 2,5 Z"
            "M 11,8 L 14,8 L 14,11 L 11,11 L 11,8 Z"
            "M 11,6 L 14,6 L 14,7 L 11,7 L 11,6 Z"
            "M 3,6 L 10,6 L 10,11 L 3,11 L 3,6 Z"
            "M 7,8 L 9,8 L 9,10 L 7,10 L 7,8 Z"
            "M 4,7 L 6,7 L 6,10 L 4,10 L 4,7 Z"
        )
    )

    expected_hole_counts = [0, 0, 1, 2]
    assert hole_counts(res) == expected_hole_counts


def test_svg_doc_from_file():
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <path d = "M 1,4 L 16,4 L 16,13 L 1,13 L 1,4 Z"/>
    </svg>
    """
    with NamedTemporaryFile("w", delete=False) as f:
        f.write(svg_src)
        f.close()
        imported = list(import_svg_document(f.name))
        assert len(imported) == 1
        try:
            os.unlink(f.name)
        except OSError:  # pragma: nocover
            pass


def test_svg_doc_path():
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <path d = "M 1,4 L 16,4 L 16,13 L 1,13 L 1,4 Z"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf))
    assert len(imported) == 1


def test_svg_doc_shapes():
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <rect width="300" height="100" fill="none"/>
    <ellipse cx="200" cy="80" rx="100" ry="50" fill="red"/>
    <circle cx="50" cy="100" r="64" fill="red"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf))
    assert len(imported) == 3


@pytest.mark.parametrize("flip_y", [False, True])
def test_svg_doc_orientation(flip_y: bool):
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <rect width="200" height="100"/>
    </svg>
    """
    buf = StringIO(svg_src)

    imported = list(import_svg_document(buf, flip_y=flip_y))

    assert len(imported) == 1
    for face in imported:
        assert isinstance(face, TopoDS_Face)
        assert as_tuple(face_normal(face)) == approx((0, 0, 1))


@pytest.mark.parametrize("flip_y", [False, True])
def test_svg_doc_transforms(flip_y: bool):
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <rect width="30" height="20" x="5" y ="15"/>
    <rect width="30" height="20" transform="translate(5 15)"/>
    <g transform="translate(5 15)">
      <rect width="30" height="20"/>
    </g>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf, flip_y=flip_y))
    assert len(imported) == 3

    def assert_bounds(s1: TopoDS_Shape, s2: TopoDS_Shape):
        b1 = bounding_box(s1)
        b2 = bounding_box(s2)
        assert as_tuple(b1.CornerMin()) == approx(as_tuple(b2.CornerMin()))
        assert as_tuple(b1.CornerMax()) == approx(as_tuple(b2.CornerMax()))

    first, *others = imported
    for other in others:
        assert_bounds(first, other)


def test_svg_doc_visibility_hidden():
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <g visibility="hidden">
    <rect width="300" height="100"/>
    </g>
    <rect width="300" height="100" display="none"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf))
    assert len(imported) == 0


def test_svg_doc_ignore_visibility():
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <g visibility="hidden">
    <rect width="300" height="100"/>
    </g>
    <rect width="300" height="100" display="none"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf, ignore_visibility=True))
    assert len(imported) == 2


@pytest.mark.parametrize("flip_y", [False, True])
def test_svg_doc_dimensions(flip_y: bool):
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg"
        width="20cm"
        height="10cm"
        viewBox="-100 -100 800 400">
    <rect width="200" height="100"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf, flip_y=flip_y))
    assert len(imported) == 1

    bb = bounding_box(imported[0])
    size = gp_Vec(bb.CornerMin(), bb.CornerMax())
    assert as_tuple(size) == approx((50, 25, 0))


def test_svg_doc_metadata_legacy():
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <rect width="5" height="10" class="blue" fill="none" stroke="#0000ff"/>
    <rect width="8" height="4" class="red" fill="#ff0000"/>
    <rect width="12" height="3"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf, metadata=ColorAndLabel.Label_by("class")))
    assert [
        (metadata.label, metadata.color_for(shape)) for shape, metadata in imported
    ] == [
        ("blue", (0, 0, 1, 1)),
        ("red", (1, 0, 0, 1)),
        ("", (0, 0, 0, 1)),
    ]


def test_svg_doc_metadata():
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <rect width="5" height="10" class="blue" fill="none" stroke="#0000ff"/>
    <rect width="8" height="4" class="red" fill="#ff0000"/>
    <rect width="12" height="3"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf, metadata=ColorAndLabel.Label_by("class")))
    assert [
        (metadata.label, metadata.color_for(shape)) for shape, metadata in imported
    ] == [
        ("blue", (0, 0, 1, 1)),
        ("red", (1, 0, 0, 1)),
        ("", (0, 0, 0, 1)),
    ]


def test_svg_nested_use_metadata():
    svg_src = """
    <svg id="svg1" viewBox="-10 -10 35 30" xmlns='http://www.w3.org/2000/svg'>
    <defs>
    <circle id="circle" cx="0" cy="0" r="5" style="opacity:1" />
    <g id="two-circles">
        <use href="#circle" id="blue_fill" transform="translate(-6 0)" fill="blue"/>
        <use href="#circle" id="red_fill" transform="translate(+6 0)" fill="red"/>
    </g>
    </defs>
    <g id="main">
    <use href="#two-circles" 
        id="white_stroke" transform="translate(5,0)" stroke="white"/>
    <use href="#two-circles"
        id="black_stroke" transform="translate(11,10)" stroke="black"/>
    </g>
    </svg>"""
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf, metadata=ColorAndLabel.Label_by("id")))

    expected = (
        (("svg1", "main", "white_stroke", "two-circles", "blue_fill"), "circle"),
        (("svg1", "main", "white_stroke", "two-circles", "red_fill"), "circle"),
        (("svg1", "main", "black_stroke", "two-circles", "blue_fill"), "circle"),
        (("svg1", "main", "black_stroke", "two-circles", "red_fill"), "circle"),
    )
    assert len(imported) == len(expected)
    for (_, metadata), (parent_labels, label) in zip(imported, expected):
        assert metadata.label == label
        assert metadata.parent_labels == parent_labels


def test_filled_but_not_a_face():
    """This path is filled shape as per the style but actually just a line segment
    so we can't make a valid face out of it"""
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <path d="M 24.281485,56.183336 H 177.76774"
        style="fill:#800000;stroke-width:0.264583"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf))
    assert len(imported) == 1
    assert isinstance(imported[0], TopoDS_Wire)


def test_filled_but_not_a_face_metadata():
    """This path is filled shape as per the style but actually just a line segment
    so we can't make a valid face out of it"""
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <path d="M 1,2 H 3"
        style="fill:red;stroke:blue"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf, metadata=ColorAndLabel))
    assert len(imported) == 1
    assert isinstance(imported[0][0], TopoDS_Wire)
    assert imported[0][1].color_for(imported[0][0]) == (0, 0, 1, 1)


def test_filled_but_not_a_face_coplanar_segments():
    """This path is filled shape as per the style but actually 2 line segments
    so we can't make a valid face out of it"""
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <path d="M 0,0 L 10,0 M 0,10 L 10,10"
        style="fill:#800000;stroke-width:0.264583"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf))
    assert len(imported) == 2
    assert isinstance(imported[0], TopoDS_Wire)
    assert isinstance(imported[1], TopoDS_Wire)


def test_filled_but_not_a_face_coaxial_segments():
    """This path is filled shape as per the style but actually 2 coaxial line segments
    so we can't make a valid face out of it"""
    svg_src = """
    <svg xmlns="http://www.w3.org/2000/svg">
    <path d="M 0,0 L 10,0 M 20,0 L 30,10"
        style="fill:#800000;stroke-width:0.264583"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf))
    assert len(imported) == 2
    assert isinstance(imported[0], TopoDS_Wire)
    assert isinstance(imported[1], TopoDS_Wire)


def test_fix_closing_lines_doc():
    """This path used to get an extremenly short closing line segment
    when converted from `svgpathelements` to `svgpathtools`.
    (fixed by converting through absolute path string)"""
    svg_src = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
    <svg xmlns="http://www.w3.org/2000/svg"
    width="22.693mm"
    height="1.6272mm"
    viewBox="0 0 22.693 1.6272"
    >
    <path d="m 7.87394,0.171449 c -0.022225,0 -0.041275,-0.007937 -0.05715,-0.023812
    -0.015875,-0.0158747 -0.023813,-0.035454 -0.023813,-0.058738 0,-0.023284
    0.00794,-0.042334 0.023813,-0.05715 0.015875,-0.01587467 0.034925,-0.023812
    0.05715,-0.023812 0.022225,0 0.041275,0.00793733 0.05715,0.023812
    0.015875,0.01481667 0.023813,0.03386667 0.023813,0.05715 0,0.02328333
    -0.00794,0.04286267 -0.023813,0.058738 -0.015875,0.0158753
    -0.034925,0.0238127 -0.05715,0.023812 z"/>
    </svg>
    """
    for e, _ in find_shapes_svg_in_document(StringIO(svg_src)):
        path = svg_element_to_path(e)
        assert path
        for segment in path:
            assert not isinstance(segment, svgelements.Line)


def test_fix_closing_lines_str():
    """`svgpathtools` would add an extremenly short closing line segment to this path.
    We want it ignored when converting to edges."""
    d = (
        "m 7.87394425193,0.171449092582 c -0.0222250120015,0 "
        "-0.0412750222885,-0.00793700428598 -0.057150030861,-0.0238120128585 c "
        "-0.0158750085725,-0.0158747085723 -0.023813012859,-0.0354540191452 "
        "-0.023813012859,-0.0587380317185 c 0,-0.0232840125734 "
        "0.0079400042876,-0.0423340228604 0.023813012859,-0.057150030861 c "
        "0.0158750085725,-0.0158746785723 0.0349250188595,-0.0238120128585 "
        "0.057150030861,-0.0238120128585 c 0.0222250120015,0 "
        "0.0412750222885,0.00793733428616 0.057150030861,0.0238120128585 c "
        "0.0158750085725,0.014816678001 0.023813012859,0.033866688288 "
        "0.023813012859,0.057150030861 c 0,0.023283342573 "
        "-0.0079400042876,0.0428626931458 -0.023813012859,0.0587380317185 c "
        "-0.0158750085725,0.0158753085727 -0.0349250188595,0.0238127128589 "
        "-0.057150030861,0.0238120128585 z"
    )
    for wire in wires_from_svg_path(d):
        for edge in topoDS_iterator(wire):
            curve = edge_to_curve(TopoDS.Edge_s(edge))
            assert curve.GetType() != GeomAbs_CurveType.GeomAbs_Line


@pytest.mark.parametrize(
    "element, curve_type, center",
    [
        (
            'circle r="40" cx="10" cy="5" transform="translate(10 0)"',
            Geom_Circle,
            (20, 5),
        ),
        (
            'circle r="40" transform="translate(10 0) scale(2 1)"',
            Geom_Ellipse,
            (10, 0),
        ),
        (
            'ellipse rx="40" ry="80" transform="scale(1 .5)"',
            Geom_Circle,
            (0, 0),
        ),
        (
            'circle r="40" transform="rotate(90) translate(10 0)"',
            Geom_Circle,
            (0, 10),
        ),
    ],
)
def test_circles_and_ellipses(
    element: str,
    curve_type: Union[Geom_Circle, Geom_Ellipse],
    center: tuple[float, float],
):
    svg_src = f"""
    <svg xmlns="http://www.w3.org/2000/svg">
        <{element} stroke="red" fill="none"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf))
    assert len(imported) == 1
    assert isinstance(imported[0], TopoDS_Wire)

    curves = [
        edge_to_curve(TopoDS.Edge_s(e)).Curve().Curve()
        for e in topoDS_iterator(imported[0])
    ]
    assert len(curves) == 1
    curve = curves[0]
    assert type(curve) is curve_type
    assert isinstance(curve, (Geom_Circle, Geom_Ellipse))
    loc = curve.Axis().Location()
    assert (loc.X(), loc.Y()) == approx(center)


@dataclass
class RoundedRect:
    w: float
    h: float
    rx: float
    ry: float

    def attrs(self):
        return f'width="{self.w}" height="{self.h}" rx="{self.rx}" ry="{self.ry}"'

    def area(self):
        rx = min(self.rx, self.w / 2)
        ry = min(self.ry, self.h / 2)
        return self.w * self.h - (2 * rx * 2 * ry - pi * rx * ry)


def rounded_rect_cases():
    rects = (
        RoundedRect(20, 10, 2, 2),
        RoundedRect(20, 10, 8, 2),
        RoundedRect(20, 10, 3, 4),
        RoundedRect(20, 10, 10, 4),
        RoundedRect(20, 10, 12, 4),
        RoundedRect(20, 10, 3, 5),
        RoundedRect(20, 10, 3, 6),
    )
    for rect in rects:
        yield rect, 0
    for rect, angle in zip(rects, cycle([10, -17, 32, -45, 90])):
        yield rect, angle


@pytest.mark.parametrize("rounded_rect, angle", rounded_rect_cases())
def test_rounded_rect(rounded_rect: RoundedRect, angle: float):
    svg_src = f"""
    <svg xmlns="http://www.w3.org/2000/svg">
    <rect {rounded_rect.attrs()} transform="rotate({angle})"/>
    </svg>
    """
    buf = StringIO(svg_src)
    imported = list(import_svg_document(buf))
    assert len(imported) == 1
    assert isinstance(imported[0], TopoDS_Face)
    assert face_area(imported[0]) == approx(rounded_rect.area())


def test_invalid_segment_logging():
    path = svgelements.Path("M 1 -2 C 3 -4 5 -6 7 -8 L 9 -10 Z")
    assert (
        str(_SegmentInPath(path[2], path))
        == "`M 7,-8 L 9,-10` segment in path `M 1,-2 C 3,-4 5,-6 7,-8 L 9,-10 Z`"
    )


def nested_squares_path(count: int, x: float = 0, y: float = 0):
    def parts():
        for s in range(1, count + 1):
            yield f"M{x-s},{y-s} H{x+s} V{y+s} H{x-s} Z"

    return " ".join(parts())


def hole_counts(maybe_faces: list[TopoDS_Face]):
    return sorted(
        len(face_inner_wires(face))
        for face in maybe_faces
        if isinstance(face, TopoDS_Face)
    )


def svg_path_tokens(path: Union[SvgPath, str]):
    def split_floats(text: str):
        i = 0
        for m in re.finditer(r"([-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?)", text):
            before = text[i : m.start(1)].strip().strip(",")
            if before:
                yield before.strip()
            yield float(m.group(1))
            i = m.end()
        end = text[i:].strip()
        if end:
            yield end

    return list(split_floats(str(path)))



from typing import Iterable, Union

import pytest
from OCP.Geom import Geom_BezierCurve, Geom_Curve, Geom_TrimmedCurve
from OCP.gp import gp_Pnt, gp_Vec
from pytest import approx, raises

from ocpsvg.ocp import (
    bezier_curve,
    circle_curve,
    closed_wire,
    curve_to_beziers,
    curve_to_bspline,
    curve_to_polyline,
    edge_from_curve,
    ellipse_curve,
    faces_from_wire_soup,
    is_wire_closed,
    segment_curve,
    wire_from_continuous_edges,
)
from tests.ocp import face_area

XY = tuple[float, float]
XYZ = tuple[float, float, float]


def Pnt(x: float, y: float, z: float = 0):
    return gp_Pnt(x, y, z)


def as_Pnt(xy: XY):
    return gp_Pnt(*xy, 0)


def as_Pnts(*xys: XY):
    return map(as_Pnt, xys)


def as_tuple(v: Union[gp_Pnt, gp_Vec]):
    return v.X(), v.Y(), v.Z()


@pytest.mark.parametrize(
    "a,b",
    [
        (Pnt(0, 0), Pnt(1, 0)),
        (Pnt(0, 0), Pnt(0, 12)),
        (Pnt(0, 0, 0), Pnt(0, 12, 34)),
    ],
)
def test_segment_curve(a: gp_Pnt, b: gp_Pnt):
    assert isinstance(segment_curve(a, b), Geom_Curve)


@pytest.mark.parametrize(
    "a,b",
    [
        (Pnt(1, 2), Pnt(1, 2)),
        (Pnt(1, 2, 3), Pnt(1, 2, 3)),
    ],
)
def test_segment_curve_invalid(a: gp_Pnt, b: gp_Pnt):
    with raises(ValueError):
        assert isinstance(segment_curve(a, b), Geom_Curve)


@pytest.mark.parametrize(
    "controls",
    [
        [Pnt(0, 0), Pnt(10, 12)],
        [Pnt(0, 1), Pnt(1, 2), Pnt(3, 4)],
        [Pnt(1, 2, 3), Pnt(0, 12, 34)],
    ],
)
def test_bezier_curve(controls: Iterable[gp_Pnt]):
    assert isinstance(bezier_curve(*controls), Geom_Curve)


@pytest.mark.parametrize(
    "controls",
    [
        [Pnt(1, 2)],
        [Pnt(i, i % 3) for i in range(26)],
    ],
)
def test_bezier_curve_invalid(controls: Iterable[gp_Pnt]):
    with raises(ValueError):
        assert isinstance(bezier_curve(*controls), Geom_Curve)


VARIOUS_CURVES: list[Geom_Curve] = [
    segment_curve(Pnt(0, 0), Pnt(1, 1)),
    bezier_curve(Pnt(0, 1), Pnt(23, 45), Pnt(67, 89)),
    bezier_curve(Pnt(0, 1), Pnt(23, 45), Pnt(66, 89), Pnt(101, 213)),
    bezier_curve(Pnt(0, 1), Pnt(23, 45), Pnt(66, 89), Pnt(101, 213), Pnt(141, 516)),
    curve_to_bspline(
        bezier_curve(Pnt(0, 1), Pnt(23, 45), Pnt(66, 89), Pnt(101, 213), Pnt(141, 516))
    ),
    circle_curve(12, clockwise=True),
    circle_curve(12, clockwise=False),
    circle_curve(12, 30, 210, clockwise=False),
    circle_curve(12, 30, 210, clockwise=True),
    ellipse_curve(12, 34, clockwise=True),
    ellipse_curve(12, 34, clockwise=False),
    ellipse_curve(12, 34, 30, 210, clockwise=False),
    ellipse_curve(12, 34, 30, 210, clockwise=True),
]


@pytest.mark.parametrize("curve", VARIOUS_CURVES)
def test_curve_to_beziers(curve: Geom_Curve):
    tol = 1e-8
    bezs = list(curve_to_beziers(curve, tolerance=tol))
    assert bezs[0].StartPoint().IsEqual(curve.Value(curve.FirstParameter()), tol)
    assert bezs[-1].EndPoint().IsEqual(curve.Value(curve.LastParameter()), tol)


@pytest.mark.parametrize("curve", VARIOUS_CURVES)
def test_trimmed_curve_to_beziers(curve: Geom_Curve):
    u, v = curve.FirstParameter(), curve.LastParameter()
    trimmed = Geom_TrimmedCurve(curve, u + (v - u) * 0.25, u + (v - u) * 0.75)
    tol = 1e-8
    bezs = list(curve_to_beziers(trimmed, tolerance=tol))
    assert bezs[0].StartPoint().IsEqual(trimmed.Value(trimmed.FirstParameter()), tol)
    assert bezs[-1].EndPoint().IsEqual(trimmed.Value(trimmed.LastParameter()), tol)


@pytest.mark.parametrize("curve", VARIOUS_CURVES)
def test_curve_to_beziers_deg3(curve: Geom_Curve):
    assert all(
        isinstance(e, Geom_BezierCurve) and e.Degree() <= 3
        for e in curve_to_beziers(curve, max_degree=3, tolerance=1e-6)
    )


@pytest.mark.parametrize("curve", VARIOUS_CURVES)
def test_curve_to_beziers_deg4(curve: Geom_Curve):
    assert all(
        isinstance(e, Geom_BezierCurve) and e.Degree() <= 4
        for e in curve_to_beziers(curve, max_degree=4, tolerance=1e-6)
    )


@pytest.mark.parametrize("curve", VARIOUS_CURVES)
def test_curve_to_polyline(curve: Geom_Curve):
    assert all(isinstance(p, gp_Pnt) for p in curve_to_polyline(curve, tolerance=1e-5))


def test_is_wire_closed():
    a = gp_Pnt(0, 0, 0)
    b = gp_Pnt(10, 0, 0)
    c = gp_Pnt(10, 10, 0)
    d = gp_Pnt(0, 10, 0)
    assert not is_wire_closed(polyline_wire(a, b, c, d))
    assert not is_wire_closed(polyline_wire(a, b, c))
    assert is_wire_closed(polyline_wire(a, b, c, a))
    assert is_wire_closed(polyline_wire(a, b, c, d, a))


def test_closed_wire():
    a = gp_Pnt(0, 0, 0)
    b = gp_Pnt(10, 0, 0)
    c = gp_Pnt(10, 10, 0)
    d = gp_Pnt(0, 10, 0)

    incomplete_loop = polyline_wire(a, b, c, d)
    incomplete_loop_fixed = closed_wire(incomplete_loop)
    assert incomplete_loop_fixed != incomplete_loop
    assert is_wire_closed(incomplete_loop_fixed)


def test_closed_wire_already_closed():
    a = gp_Pnt(0, 0, 0)
    b = gp_Pnt(10, 0, 0)
    c = gp_Pnt(10, 10, 0)

    loop = polyline_wire(a, b, c, a)
    assert closed_wire(loop) == loop


def test_closed_wire_single_segment_noop():
    a = gp_Pnt(0, 0, 0)
    b = gp_Pnt(10, 0, 0)
    incomplete_loop = polyline_wire(a, b)
    incomplete_loop_fixed = closed_wire(incomplete_loop)
    assert incomplete_loop_fixed == incomplete_loop


def test_face_from_wire_soup_winding():
    a = gp_Pnt(0, 0, 0)
    b = gp_Pnt(10, 0, 0)
    c = gp_Pnt(10, 10, 0)
    d = gp_Pnt(0, 10, 0)

    e = gp_Pnt(2, 2, 0)
    f = gp_Pnt(7, 2, 0)
    g = gp_Pnt(7, 7, 0)
    h = gp_Pnt(2, 7, 0)

    def faces_area_from_rings(*rings: list[gp_Pnt]):
        return [
            face_area(f)
            for f in faces_from_wire_soup(polyline_wire(*r, r[0]) for r in rings)
        ]

    assert faces_area_from_rings([a, b, c, d]) == approx([100.0])
    assert faces_area_from_rings([d, c, b, a]) == approx([100.0])
    assert faces_area_from_rings([a, b, c, d], [e, f, g, h]) == approx([75.0])
    assert faces_area_from_rings([a, b, c, d], [h, g, f, e]) == approx([75.0])
    assert faces_area_from_rings([d, c, b, a], [e, f, g, h]) == approx([75.0])
    assert faces_area_from_rings([d, c, b, a], [h, g, f, e]) == approx([75.0])


def polyline_wire(*points: gp_Pnt):
    return wire_from_continuous_edges(
        edge_from_curve(segment_curve(p, q)) for p, q in zip(points, points[1:])
    )



from __future__ import annotations

from dataclasses import dataclass
from itertools import product
from pathlib import Path
from typing import (
    Any,
    BinaryIO,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Union,
)
from xml.etree import ElementTree as ET

from OCP.Bnd import Bnd_Box
from OCP.BRepBuilderAPI import BRepBuilderAPI_Transform
from OCP.gp import gp_Ax2, gp_Dir, gp_Pnt, gp_Trsf, gp_Vec
from OCP.HLRAlgo import HLRAlgo_Projector
from OCP.HLRBRep import HLRBRep_Algo, HLRBRep_HLRToShape, HLRBRep_TypeOfResultingEdge
from OCP.TopoDS import TopoDS, TopoDS_Edge, TopoDS_Shape

from .ocp import bounding_box, topoDS_iterator
from .svg import edge_to_svg_path, format_svg

HlrEdgeTypeName = Literal["undefined", "isoline", "sewn", "smooth", "sharp", "outline"]

EDGE_TYPES_NAMES: dict[HLRBRep_TypeOfResultingEdge, HlrEdgeTypeName] = {
    HLRBRep_TypeOfResultingEdge.HLRBRep_Undefined: "undefined",
    HLRBRep_TypeOfResultingEdge.HLRBRep_IsoLine: "isoline",
    HLRBRep_TypeOfResultingEdge.HLRBRep_RgNLine: "sewn",
    HLRBRep_TypeOfResultingEdge.HLRBRep_Rg1Line: "smooth",
    HLRBRep_TypeOfResultingEdge.HLRBRep_Sharp: "sharp",
    HLRBRep_TypeOfResultingEdge.HLRBRep_OutLine: "outline",
}


Padding = Union[
    float,
    tuple[float],
    tuple[float, float],
    tuple[float, float, float],
    tuple[float, float, float, float],
]
CssStyle = Mapping[str, Mapping[str, Any]]
RGB = tuple[float, float, float]
EdgeColor = Union[RGB, tuple[RGB, RGB]]


PRIMARY_EDGES = "sharp", "outline"
SECONDARY_EDGES = "undefined", "isoline", "sewn", "smooth"


@dataclass
class HlrEdge:
    type: HLRBRep_TypeOfResultingEdge
    is_hidden: bool
    projected_edge: TopoDS_Edge
    edge_in_3d: TopoDS_Edge
    shape_index: int

    @property
    def type_name(self) -> HlrEdgeTypeName:
        return EDGE_TYPES_NAMES.get(self.type, "undefined")


@dataclass
class HiddenLineRenderer:
    projector: HLRAlgo_Projector

    EDGE_TYPES = [
        # ignore HLRBRep_TypeOfResultingEdge.HLRBRep_Undefined
        HLRBRep_TypeOfResultingEdge.HLRBRep_IsoLine,
        HLRBRep_TypeOfResultingEdge.HLRBRep_RgNLine,
        HLRBRep_TypeOfResultingEdge.HLRBRep_Rg1Line,
        HLRBRep_TypeOfResultingEdge.HLRBRep_Sharp,
        HLRBRep_TypeOfResultingEdge.HLRBRep_OutLine,
    ]

    @classmethod
    def Orthographic(
        cls,
        camera_direction: gp_Dir = gp_Dir(-1, +1, -1),
        camera_up: gp_Dir = gp_Dir(0, 0, 1),
    ):
        camera_ax = gp_Ax2(gp_Pnt(0, 0, 0), camera_direction.Reversed())
        camera_ax.SetYDirection(camera_up)
        projector = HLRAlgo_Projector(camera_ax)
        return cls(projector)

    @classmethod
    def Perspective(
        cls,
        camera_position: gp_Pnt,
        camera_focus: gp_Pnt = gp_Pnt(0, 0, 0),
        focal_length: float = 400,
        camera_up: gp_Dir = gp_Dir(0, 0, 1),
    ):
        camera_ax = gp_Ax2(camera_focus, gp_Dir(gp_Vec(camera_focus, camera_position)))
        camera_ax.SetYDirection(camera_up)
        projector = HLRAlgo_Projector(camera_ax, focal_length)
        return cls(projector)

    def __call__(
        self, shapes: Iterable[TopoDS_Shape], with_hidden: bool = True
    ) -> HiddenLineRender:
        def rough_z(e: HlrEdge):
            trsf = self.projector.FullTransformation()
            pseudo_projected = BRepBuilderAPI_Transform(e.edge_in_3d, trsf).Shape()
            return bounding_box(pseudo_projected).CornerMax().Z()

        return HiddenLineRender(
            sorted(self.compute_edges(shapes, with_hidden=with_hidden), key=rough_z),
            self.projector,
        )

    def compute_edges(self, shapes: Iterable[TopoDS_Shape], with_hidden: bool = True):
        hlr = HLRBRep_Algo()
        for shape in shapes:
            hlr.Add(shape)

        hlr.Projector(self.projector)
        hlr.Update()
        hlr.Hide()

        visibilities = (False, True) if with_hidden else (True,)

        for shape_index in range(hlr.NbShapes()):
            hlr.Select(shape_index + 1)
            hlr_shapes = HLRBRep_HLRToShape(hlr)

            for is_visible, type in product(visibilities, self.EDGE_TYPES):
                edges_in_2d = hlr_shapes.CompoundOfEdges(type, is_visible, In3d=False)
                edges_in_3d = hlr_shapes.CompoundOfEdges(type, is_visible, In3d=True)

                for edge_in_2d, edge_in_3d in zip(
                    topoDS_iterator(edges_in_2d), topoDS_iterator(edges_in_3d)
                ):
                    yield HlrEdge(
                        type=type,
                        is_hidden=not is_visible,
                        projected_edge=TopoDS.Edge_s(edge_in_2d),
                        edge_in_3d=TopoDS.Edge_s(edge_in_3d),
                        shape_index=shape_index,
                    )


class HiddenLineRender(Iterable[HlrEdge]):
    def __init__(
        self, edges: Iterable[HlrEdge], projector: Optional[HLRAlgo_Projector] = None
    ) -> None:
        self.edges = list(edges)
        self.projector = projector

    def __iter__(self) -> Iterator[HlrEdge]:
        return iter(self.edges)

    def bounds(self) -> Bnd_Box:
        return bounding_box(e.projected_edge for e in self)

    def to_svg(
        self,
        width: Optional[float] = 512,
        height: Optional[float] = None,
        *,
        padding: Padding = 16,
        css_style: Optional[CssStyle] = None,
        background: Optional[bool] = None,
        tolerance: float = 1e-6,
        float_format: str = "g",
    ) -> ET.ElementTree:
        (x0, y0, x1, y1), scale, (W, H) = _viewbox_and_scale(
            self.bounds(),
            width=width,
            height=height,
            padding=padding,
        )
        ty = -(y1 + y0) / scale

        def fmt(v: float):
            return v.__format__(float_format)

        svg = ET.Element("svg", xmlns="http://www.w3.org/2000/svg")
        svg.attrib["viewBox"] = f"{fmt(x0)} {fmt(y0)} {fmt(x1-x0)} {fmt(y1-y0)}"
        if width:
            svg.attrib["width"] = f"{width}"
        if height:
            svg.attrib["height"] = f"{height}"

        if css_style is None:
            css_style = basic_style()
        if css_style:
            ET.SubElement(svg, "style").text = "\n".join(css_style_to_lines(css_style))

        if background is None:
            background = "#background" in css_style
        if background:
            pad = 1
            ET.SubElement(
                svg,
                "rect",
                id="background",
                x=f"{ (x0+x1)/2 - W/2 - pad }",
                y=f"{ (y0+y1)/2 - H/2 - pad }",
                width=f"{ W + 2*pad  }",
                height=f"{ H + 2*pad  }",
            )

        # can't scale/mirror `BRep_CurveOnSurface` edges :( so we do that in the SVG
        # and use `vector-effect: non-scaling-stroke` style to keep intended line witdhs
        transformed_group = ET.SubElement(
            svg,
            "g",
            transform=f"scale({fmt(scale)} {fmt(-scale)}) translate(0 {fmt(ty)})",
        )

        for i, edge in enumerate(self):
            classnames = [f"s{edge.shape_index}", edge.type_name]
            if edge.is_hidden:
                classnames.append("hidden")

            attrs = {
                "id": f"e{i}",
                "d": format_svg(
                    edge_to_svg_path(edge.projected_edge, tolerance=tolerance),
                    float_format=float_format,
                ),
                "class": " ".join(classnames),
            }
            ET.SubElement(transformed_group, "path", attrs)

        return ET.ElementTree(svg)


def write_svg(
    svg: ET.ElementTree,
    f: Union[str, Path, BinaryIO],
    indent: bool = True,
):
    if indent:
        ET.indent(svg)
    svg.write(f, xml_declaration=True, encoding="utf-8")


def basic_style(
    color: EdgeColor = ((0, 0, 0), (0.3, 0.3, 0.3)),
    linewidth: float = 1,
    secondary_linewidth: Optional[float] = None,
    background_color: Optional[RGB] = None,
    shape_colors: Optional[Mapping[int, EdgeColor]] = None,
) -> CssStyle:
    if secondary_linewidth is None:
        secondary_linewidth = linewidth * 0.75
    color, hidden_color = _edge_color_pair(color)

    style = {
        "path": {
            "fill": "none",
            "stroke": hexcolor(color),
            "stroke-width": linewidth,
            "stroke-linecap": "round",
            "stroke-linejoin": "round",
            "vector-effect": "non-scaling-stroke",
        },
        ", ".join(f".{t}" for t in SECONDARY_EDGES): {
            "stroke-width": secondary_linewidth,
        },
        ".hidden": {
            "stroke": hexcolor(hidden_color),
            "stroke-dasharray": f"{1.5*linewidth} {2.5*linewidth}",
        },
    }

    if background_color:
        style["#background"] = {"fill": hexcolor(background_color)}

    if shape_colors:
        for i, color in shape_colors.items():
            color, hidden_color = _edge_color_pair(color)
            style[f".s{i}"] = {"stroke": hexcolor(color)}
            style[f".s{i}.hidden"] = {"stroke": hexcolor(hidden_color)}

    return style


def hexcolor(rgb: tuple[float, float, float]):
    r, g, b = [min(max(int(round(v * 255)), 0), 255) for v in rgb]
    return f"#{r:02x}{g:02x}{b:02x}"


def _edge_color_pair(rgb_or_pair: EdgeColor):
    if len(rgb_or_pair) == 2:
        return rgb_or_pair
    else:
        return rgb_or_pair, rgb_or_pair


def _viewbox_and_scale(
    bbox: Bnd_Box,
    width: Optional[float] = None,
    height: Optional[float] = None,
    *,
    padding: Padding = 0,
):
    padding_t, padding_r, padding_b, padding_l = _normalize_padding(padding)

    c0 = bbox.CornerMin()
    c1 = bbox.CornerMax()
    bounds_width = c1.X() - c0.X()
    bounds_height = c1.Y() - c0.Y()

    if width and height:
        scale = min(
            (width - padding_r - padding_l) / bounds_width,
            (height - padding_b - padding_t) / bounds_height,
        )
    elif width:
        scale = (width - padding_r - padding_l) / bounds_width
        height = bounds_height * scale + padding_b + padding_t
    elif height:
        scale = (height - padding_b - padding_t) / bounds_height
        width = bounds_width * scale + padding_r + padding_l
    else:
        scale = 1
        width = bounds_width + padding_r + padding_l
        height = bounds_height + padding_b + padding_t

    trsf = gp_Trsf()
    trsf.SetScale(gp_Pnt(), scale)
    c0.Transform(trsf)
    c1.Transform(trsf)

    x0 = c0.X() - padding_l
    y0 = c0.Y() - padding_b
    x1 = c1.X() + padding_r
    y1 = c1.Y() + padding_t

    return (x0, y0, x1, y1), scale, (width, height)


def _normalize_padding(p: Padding) -> tuple[float, float, float, float]:
    if isinstance(p, Sequence):
        if len(p) == 1:
            return p[0], p[0], p[0], p[0]
        elif len(p) == 2:
            return p[0], p[1], p[0], p[1]
        elif len(p) == 3:
            return p[0], p[1], p[2], p[1]
        else:
            return p[0], p[1], p[2], p[3]
    else:
        return p, p, p, p


def css_style_to_lines(style: CssStyle):
    for selector, props in style.items():
        yield selector + " {"
        for prop, value in props.items():
            yield f"{prop}: { value };"
        yield "}"



import logging
from itertools import chain
from math import radians
from typing import Iterable, Iterator, Optional, Union, cast

from OCP.Bnd import Bnd_Box
from OCP.BRepAdaptor import BRepAdaptor_Curve
from OCP.BRepBndLib import BRepBndLib
from OCP.BRepBuilderAPI import BRepBuilderAPI_MakeEdge, BRepBuilderAPI_MakeFace
from OCP.BRepFeat import BRepFeat
from OCP.BRepLib import BRepLib_FindSurface
from OCP.BRepTools import BRepTools
from OCP.GC import (
    GC_MakeArcOfCircle,
    GC_MakeArcOfEllipse,
    GC_MakeCircle,
    GC_MakeEllipse,
    GC_MakeSegment,
)
from OCP.GCPnts import GCPnts_QuasiUniformDeflection
from OCP.Geom import (
    Geom_BezierCurve,
    Geom_BSplineCurve,
    Geom_Circle,
    Geom_Curve,
    Geom_Ellipse,
    Geom_TrimmedCurve,
)
from OCP.GeomAbs import GeomAbs_CurveType, GeomAbs_Shape
from OCP.GeomAdaptor import GeomAdaptor_Curve
from OCP.GeomConvert import (
    GeomConvert,
    GeomConvert_ApproxCurve,
    GeomConvert_BSplineCurveToBezierCurve,
)
from OCP.gp import (
    gp_Ax1,
    gp_Ax2,
    gp_Circ,
    gp_Dir,
    gp_Elips,
    gp_Pnt,
    gp_Trsf,
    gp_Vec,
)
from OCP.ShapeAnalysis import ShapeAnalysis_Wire
from OCP.ShapeExtend import ShapeExtend_WireData
from OCP.ShapeFix import ShapeFix_Face
from OCP.Standard import Standard_Failure
from OCP.StdFail import StdFail_NotDone
from OCP.TColgp import TColgp_Array1OfPnt
from OCP.TopoDS import (
    TopoDS,
    TopoDS_Builder,
    TopoDS_Compound,
    TopoDS_Edge,
    TopoDS_Face,
    TopoDS_Iterator,
    TopoDS_Shape,
    TopoDS_Wire,
)

logger = logging.getLogger(__name__)


_TOLERANCE = 1e-8


#### shapes


def make_compound(shapes: Iterable[TopoDS_Shape]) -> TopoDS_Compound:
    compound = TopoDS_Compound()
    comp_builder = TopoDS_Builder()
    comp_builder.MakeCompound(compound)

    for shape in shapes:
        comp_builder.Add(compound, shape)

    return compound


def bounding_box(
    shape_or_shapes: Union[TopoDS_Shape, Iterable[TopoDS_Shape]]
) -> Bnd_Box:
    bbox = Bnd_Box()
    for shape in (
        [shape_or_shapes]
        if isinstance(shape_or_shapes, TopoDS_Shape)
        else shape_or_shapes
    ):
        BRepBndLib.AddOptimal_s(shape, bbox)
    return bbox


def topoDS_iterator(
    shape: TopoDS_Shape, with_orientation: bool = True, with_location: bool = True
) -> Iterator[TopoDS_Shape]:
    iterator = TopoDS_Iterator(shape, with_orientation, with_location)
    while iterator.More():
        yield iterator.Value()
        iterator.Next()


#### faces


def face_outer_wire(face: TopoDS_Face) -> TopoDS_Wire:
    """Find the outer wire of a face."""
    return BRepTools.OuterWire_s(face)


def face_inner_wires(face: TopoDS_Face) -> list[TopoDS_Wire]:
    """Find the inner wires of a face."""
    outer = face_outer_wire(face)
    return [TopoDS.Wire_s(w) for w in topoDS_iterator(face) if not w.IsSame(outer)]


def face_from_wires(
    outer_wire: TopoDS_Wire, inner_wires: Optional[Iterable[TopoDS_Wire]] = None
) -> TopoDS_Face:
    """Make a face from an outer wire and optional inner wire(s)."""
    face_builder = BRepBuilderAPI_MakeFace(outer_wire, True)
    if inner_wires:
        for inner_wire in inner_wires:
            face_builder.Add(inner_wire)

    face_fix = ShapeFix_Face(face_builder.Face())
    face_fix.FixOrientation()
    return face_fix.Face()


class InvalidWiresForFace(ValueError):
    pass


def faces_from_wire_soup(wires: Iterable[TopoDS_Wire]) -> Iterable[TopoDS_Face]:
    """Make faces from unorganized, possibly nested (but non-intersecting) wires."""

    wires = list(wires)

    for wire in wires:
        if not are_wires_coplanar([wire]):
            raise InvalidWiresForFace("invalid wire")

    if not are_wires_coplanar(wires):
        raise InvalidWiresForFace("wires are not coplanar")

    def fix_wires():
        for wire in wires:
            # TODO split self intersecting wires?
            yield face_outer_wire(face_from_wires(wire))

    faces = [BRepBuilderAPI_MakeFace(wire, True).Face() for wire in fix_wires()]

    if len(faces) < 2:
        yield from faces
        return

    included_in: dict[int, set[int]] = {}
    for i, face_i in enumerate(faces):
        for j, face_j in enumerate(faces):
            if i != j and BRepFeat.IsInside_s(face_i, face_j):
                included_in.setdefault(i, set()).add(j)

    outers_and_inners: dict[int, tuple[list[TopoDS_Wire], list[TopoDS_Wire]]] = {}

    for i in range(len(faces)):
        ancestors = included_in.get(i, set())
        if len(ancestors) % 2:  # odd depth: inner ring
            parent_i = max(ancestors, key=lambda i: len(included_in.get(i, set())))
            _, inners = outers_and_inners.setdefault(parent_i, ([], []))
            inners.append(face_outer_wire(faces[i]))
        else:  # even depth: outer ring
            outers, _ = outers_and_inners.setdefault(i, ([], []))
            outers.append(face_outer_wire(faces[i]))

    for outers, inners in outers_and_inners.values():
        if len(outers) == 1:
            yield face_from_wires(outers[0], inners)
        else:  # pragma: nocover
            # shouldn't ever get here
            # but yield everything as simple faces just in case
            logger.warning("invalid nesting (found %d outer wires)", len(outers))
            for path in chain(outers, inners):
                yield face_from_wires(path)


#### wires


def is_wire_closed(wire: TopoDS_Wire) -> bool:
    """Check whether a wire is closed."""
    a = ShapeAnalysis_Wire(wire, BRepBuilderAPI_MakeFace(wire).Face(), _TOLERANCE)
    return a.CheckClosed()


def are_wires_coplanar(wires: Iterable[TopoDS_Wire]) -> bool:
    """Check whether wires are coplanar."""
    wires = list(wires)
    return (
        not wires or BRepLib_FindSurface(make_compound(wires), OnlyPlane=True).Found()
    )


def wire_from_continuous_edges(
    edges: Iterable[TopoDS_Edge], *, closed: bool = False
) -> TopoDS_Wire:
    """Make a single wire from known-continuous edges;
    with no reordering nor any checking."""
    extend = ShapeExtend_WireData(TopoDS_Wire())
    for edge in edges:
        extend.AddOriented(edge, 0)

    wire = extend.Wire()
    return closed_wire(wire) if closed else wire


def closed_wire(wire: TopoDS_Wire) -> TopoDS_Wire:
    """Ensure a wire is closed.
    - if already closed return it untouched.
    - otherwise return a new wire with a closing segment appended."""

    if is_wire_closed(wire):
        return wire

    it = topoDS_iterator(wire)
    try:
        first_edge = next(it)
        last_edge = next(it)
        while True:
            try:
                last_edge = next(it)
            except StopIteration:
                break
    except StopIteration:
        # wire has fewer that 2 edges
        return wire

    adaptor = BRepAdaptor_Curve(TopoDS.Edge_s(first_edge))
    start = adaptor.Value(adaptor.FirstParameter())

    adaptor = BRepAdaptor_Curve(TopoDS.Edge_s(last_edge))
    end = adaptor.Value(adaptor.LastParameter())

    if not start.IsEqual(end, _TOLERANCE):
        extend = ShapeExtend_WireData()
        extend.AddOriented(wire, 0)
        extend.AddOriented(edge_from_curve(segment_curve(end, start)), 0)
        wire = extend.Wire()

    return wire


#### edges


def edge_from_curve(curve: Geom_Curve) -> TopoDS_Edge:
    return BRepBuilderAPI_MakeEdge(curve).Edge()


def edge_to_curve(edge: TopoDS_Edge) -> BRepAdaptor_Curve:
    return BRepAdaptor_Curve(edge)


#### curves


BEZIER_MAX_DEGREE = Geom_BezierCurve.MaxDegree_s()

CurveOrAdaptor = Union[Geom_Curve, GeomAdaptor_Curve, BRepAdaptor_Curve]


def segment_curve(start: gp_Pnt, end: gp_Pnt) -> Geom_TrimmedCurve:
    try:
        return GC_MakeSegment(start, end).Value()
    except (StdFail_NotDone, Standard_Failure) as e:
        raise ValueError(f"could not make segment curve from {start}, {end}", e)


def bezier_curve(*controls: gp_Pnt) -> Geom_BezierCurve:
    n = len(controls)
    if not 2 <= n <= BEZIER_MAX_DEGREE:
        raise ValueError(
            f"bezier curve must have between 2 and {BEZIER_MAX_DEGREE} control points"
        )

    poles = TColgp_Array1OfPnt(1, n)
    for i, control in enumerate(controls, 1):
        poles.SetValue(i, control)

    return Geom_BezierCurve(poles)


def circle_curve(
    radius: float,
    start_angle: float = 360,
    end_angle: float = 360,
    *,
    clockwise: bool = False,
    center: gp_Pnt = gp_Pnt(0, 0, 0),
    normal: gp_Dir = gp_Dir(0, 0, 1),
) -> Union[Geom_Circle, Geom_TrimmedCurve]:
    circle_gp = gp_Circ(gp_Ax2(gp_Pnt(), normal), radius)

    if start_angle == end_angle:
        circle = GC_MakeCircle(circle_gp).Value()
    else:
        circle = GC_MakeArcOfCircle(
            circle_gp, radians(start_angle), radians(end_angle), clockwise
        ).Value()

    trsf = gp_Trsf()
    trsf.SetTranslation(gp_Vec(center.XYZ()))
    circle.Transform(trsf)

    return circle


def ellipse_curve(
    major_radius: float,
    minor_radius: float,
    start_angle: float = 360,
    end_angle: float = 360,
    *,
    clockwise: bool = False,
    rotation: float = 0,
    center: gp_Pnt = gp_Pnt(0, 0, 0),
    normal: gp_Dir = gp_Dir(0, 0, 1),
) -> Union[Geom_Ellipse, Geom_TrimmedCurve]:
    if minor_radius > major_radius:
        major_radius, minor_radius = minor_radius, major_radius
        angle_adjustment = 90
    else:
        angle_adjustment = 0

    ellipse_gp = gp_Elips(gp_Ax2(gp_Pnt(), normal), major_radius, minor_radius)
    ellipse_gp.Rotate(gp_Ax1(), radians(angle_adjustment))

    if start_angle == end_angle:
        ellipse = GC_MakeEllipse(ellipse_gp).Value()
    else:
        ellipse = GC_MakeArcOfEllipse(
            ellipse_gp,
            radians(start_angle - angle_adjustment),
            radians(end_angle - angle_adjustment),
            clockwise,
        ).Value()

    ellipse.Rotate(gp_Ax1(gp_Pnt(), normal), radians(rotation))
    ellipse.Translate(gp_Vec(center.XYZ()))

    return ellipse


def curve_to_bspline(curve: Geom_Curve) -> Geom_BSplineCurve:
    return GeomConvert.CurveToBSplineCurve_s(curve)  # type: ignore


def curve_to_beziers(
    curve_or_adaptor: CurveOrAdaptor,
    *,
    tolerance: float,
    max_degree: int = 3,
    max_segments: int = 100,
) -> Iterator[Geom_BezierCurve]:
    curve, adaptor = curve_and_adaptor(curve_or_adaptor)
    curve_type = adaptor.GetType()

    if curve_type == GeomAbs_CurveType.GeomAbs_Line:
        start = adaptor.Value(adaptor.FirstParameter())
        end = adaptor.Value(adaptor.LastParameter())
        yield bezier_curve(start, end)

    elif curve_type == GeomAbs_CurveType.GeomAbs_BezierCurve:
        bezier = cast(Geom_BezierCurve, adaptor.Bezier().Copy())
        bezier.Segment(adaptor.FirstParameter(), adaptor.LastParameter())
        if bezier.Degree() > max_degree or bezier.IsRational():
            yield from bspline_to_beziers(
                curve_to_bspline(bezier),
                max_degree=max_degree,
                max_segments=max_segments,
                tolerance=tolerance,
            )
        else:
            yield bezier

    elif curve_type == GeomAbs_CurveType.GeomAbs_BSplineCurve:
        bspline = cast(Geom_BSplineCurve, adaptor.BSpline().Copy())
        bspline.Segment(adaptor.FirstParameter(), adaptor.LastParameter())
        yield from bspline_to_beziers(
            bspline,
            max_degree=max_degree,
            max_segments=max_segments,
            tolerance=tolerance,
        )

    else:
        yield from bspline_to_beziers(
            curve_to_bspline(curve),
            max_degree=max_degree,
            max_segments=max_segments,
            tolerance=tolerance,
        )


def bspline_to_beziers(
    bspline: Geom_BSplineCurve,
    *,
    tolerance: float,
    max_degree: int = 3,
    max_segments: int = 100,
) -> Iterator[Geom_BezierCurve]:
    bspline = cast(Geom_BSplineCurve, bspline.Copy())
    bspline.Segment(bspline.FirstParameter(), bspline.LastParameter())
    if bspline.Degree() > 3 or bspline.IsRational():
        approx = GeomConvert_ApproxCurve(
            bspline,
            tolerance,
            GeomAbs_Shape.GeomAbs_C0,
            MaxSegments=max_segments,
            MaxDegree=max_degree,
        )
        if approx.IsDone() and approx.HasResult():
            bspline = approx.Curve()
        else:
            raise ValueError(f"could not approximate b-spline {bspline}")

    bez_convert = GeomConvert_BSplineCurveToBezierCurve(bspline)
    for i in range(bez_convert.NbArcs()):
        yield bez_convert.Arc(i + 1)


def curve_to_polyline(
    curve_or_adaptor: CurveOrAdaptor,
    *,
    tolerance: float,
) -> Iterator[gp_Pnt]:
    _curve, adaptor = curve_and_adaptor(curve_or_adaptor)

    start = adaptor.FirstParameter()
    end = adaptor.LastParameter()

    curve_type = adaptor.GetType()
    if curve_type == GeomAbs_CurveType.GeomAbs_Line:
        yield adaptor.Value(start)
        yield adaptor.Value(end)
    else:
        points = GCPnts_QuasiUniformDeflection(adaptor, tolerance, start, end)
        if points.IsDone():
            for i in range(points.NbPoints()):
                yield points.Value(i + 1)
        else:
            raise ValueError("could not convert to polyline")


def curve_and_adaptor(
    curve_or_adaptor: CurveOrAdaptor,
) -> tuple[Geom_Curve, Union[GeomAdaptor_Curve, BRepAdaptor_Curve]]:
    if isinstance(curve_or_adaptor, Geom_Curve):
        return curve_or_adaptor, GeomAdaptor_Curve(curve_or_adaptor)
    elif isinstance(curve_or_adaptor, GeomAdaptor_Curve):
        return curve_or_adaptor.Curve(), curve_or_adaptor
    elif isinstance(curve_or_adaptor, BRepAdaptor_Curve):
        return curve_or_adaptor.Curve().Curve(), curve_or_adaptor
    else:
        raise TypeError()
